// glwrapperView.cpp : implementation of the CGlwrapperView class
//

#include "stdafx.h"
#include "glwrapper.h"

#include "glwrapperDoc.h"
#include "glwrapperView.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


CGlwrapperView * CGlwrapperView::m_glOwner = NULL;


static BYTE threeto8[8] =  { 0, 0111>>1, 0222>>1, 0333>>1, 0444>>1, 0555>>1, 0666>>1, 0377 };
static BYTE twoto8[4]   =  { 0, 0x55, 0xaa, 0xff };
static BYTE oneto8[2]   =  { 0, 255 };
static int  defaultOverride[13] = { 0, 3, 24, 27, 64, 67, 88, 173, 181, 236, 247, 164, 91 };

static PALETTEENTRY defaultPalEntry[20] = 
{
	{ 0,   0,   0,    0 },
	{ 0x80,0,   0,    0 },
	{ 0,   0x80,0,    0 },
	{ 0x80,0x80,0,    0 },
	{ 0,   0,   0x80, 0 },
	{ 0x80,0,   0x80, 0 },
	{ 0,   0x80,0x80, 0 },
	{ 0xC0,0xC0,0xC0, 0 },

	{ 192, 220, 192,  0 },
	{ 166, 202, 240,  0 },
	{ 255, 251, 240,  0 },
	{ 160, 160, 164,  0 },

	{ 0x80,0x80,0x80, 0 },
	{ 0xFF,0,   0,    0 },
	{ 0,   0xFF,0,    0 },
	{ 0xFF,0xFF,0,    0 },
	{ 0,   0,   0xFF, 0 },
	{ 0xFF,0,   0xFF, 0 },
	{ 0,   0xFF,0xFF, 0 },
	{ 0xFF,0xFF,0xFF, 0 }
};


	// Funzioni statiche di supporto

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift);
static BOOL SetupPixelFormat( CDC * pDC );
static BOOL SetupPrintPixelFormat( CDC * pDC );
static BOOL CreateRGBPalette( CDC * pDC, CPalette & cPalette );
static void InitGL( void );

static unsigned char ComponentFromIndex(int i, UINT nbits, UINT shift)
{
    unsigned char val;

    val = (unsigned char) (i >> shift);
    switch (nbits) 
	{

    case 1:
        val &= 0x1;
        return oneto8[val];
    case 2:
        val &= 0x3;
        return twoto8[val];
    case 3:
        val &= 0x7;
        return threeto8[val];

    default:
        return 0;
    }
}



static BOOL SetupPixelFormat( CDC * pDC )
{
    static PIXELFORMATDESCRIPTOR pfd = 
	{
        sizeof(PIXELFORMATDESCRIPTOR),  // size of this pfd
        1,                              // version number
        PFD_DRAW_TO_WINDOW |            // support window
        PFD_SUPPORT_OPENGL |            // support OpenGL
        PFD_DOUBLEBUFFER,               // double buffered
        PFD_TYPE_RGBA,                  // RGBA type
        24,                             // 24-bit color depth
        0, 0, 0, 0, 0, 0,               // color bits ignored
        0,                              // no alpha buffer
        0,                              // shift bit ignored
        0,                              // no accumulation buffer
        0, 0, 0, 0,                     // accum bits ignored
        32,                             // 32-bit z-buffer
		// STENCIL
        0,                              // no stencil buffer
        0,                              // no auxiliary buffer
        PFD_MAIN_PLANE,                 // main layer
        0,                              // reserved
        0, 0, 0                         // layer masks ignored
    };
    int pixelformat;

    if ( (pixelformat = ChoosePixelFormat(pDC->GetSafeHdc(), &pfd)) == 0 )
    {
        AfxMessageBox("ChoosePixelFormat failed");
	    return FALSE;
    }

    if (SetPixelFormat(pDC->GetSafeHdc(), pixelformat, &pfd) == FALSE)
    {
        AfxMessageBox("SetPixelFormat failed");
	    return FALSE;
    }

    return TRUE;
}



static BOOL CreateRGBPalette( CDC * pDC, CPalette & cPalette )
{
    PIXELFORMATDESCRIPTOR pfd;
    LOGPALETTE *pPal;
    int n, i;
 
    n = ::GetPixelFormat(pDC->GetSafeHdc());
    ::DescribePixelFormat(pDC->GetSafeHdc(), n, sizeof(pfd), &pfd);

    if (pfd.dwFlags & PFD_NEED_PALETTE)
    {
        n = 1 << pfd.cColorBits;
        pPal = (PLOGPALETTE) new char[sizeof(LOGPALETTE) + n * sizeof(PALETTEENTRY)];

        ASSERT(pPal != NULL);

        pPal->palVersion = 0x300;
        pPal->palNumEntries = (WORD)n;
        for (i=0; i<n; i++)
        {
            pPal->palPalEntry[i].peRed =
                    ComponentFromIndex(i, pfd.cRedBits, pfd.cRedShift);
            pPal->palPalEntry[i].peGreen =
                    ComponentFromIndex(i, pfd.cGreenBits, pfd.cGreenShift);
            pPal->palPalEntry[i].peBlue =
                    ComponentFromIndex(i, pfd.cBlueBits, pfd.cBlueShift);
            pPal->palPalEntry[i].peFlags = 0;
        }

        /* fix up the palette to include the default GDI palette */
        if ((pfd.cColorBits == 8)                           &&
            (pfd.cRedBits   == 3) && (pfd.cRedShift   == 0) &&
            (pfd.cGreenBits == 3) && (pfd.cGreenShift == 3) &&
            (pfd.cBlueBits  == 2) && (pfd.cBlueShift  == 6)
           )
        {
			for (i = 1 ; i <= 12 ; i++)
                pPal->palPalEntry[defaultOverride[i]] = defaultPalEntry[i];
        }

        cPalette.CreatePalette(pPal);
        delete pPal;

		return TRUE;
    }
	return FALSE;
}


static void InitGL( void )
{
    glClearDepth(1.0f);
	glShadeModel( GL_SMOOTH );
    glEnable(GL_DEPTH_TEST);
	glEnable(GL_COLOR_MATERIAL);
	glDisable(GL_CULL_FACE);
	glDisable(GL_AUTO_NORMAL);
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_LINE_SMOOTH);
	glDisable(GL_POINT_SMOOTH);
	glDisable(GL_POLYGON_SMOOTH);
	glEnable(GL_NORMALIZE);
	glDisable(GL_FOG);
	glLineWidth(1.0f);
	glEnable(GL_TEXTURE_2D);

	/* STENCIL
	glEnable( GL_STENCIL );
	glClearStencil( 0 );
	*/

		// Settaggi (eventuale) texture
	const GLfloat bcolor[4]={0.5f,0.5f,0.5f,1.0f};
	if(/*m_bTextMode*/0)
	{
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR );
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR );
	}
	else
	{
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST );
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_NEAREST );
	}
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
	glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);
	//glTexParameterfv(GL_TEXTURE_2D,GL_TEXTURE_BORDER_COLOR,bcolor);

	glPixelStorei( GL_PACK_ROW_LENGTH, 0);
	glPixelStorei( GL_PACK_ALIGNMENT, 1);


		// Settaggi luci
    float fv4[4];
    int iv1[1];

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

	fv4[0] = 0.1f; fv4[1] = 0.1f; fv4[2] = 0.1f; fv4[3] = 1.0f;
    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, fv4);
    
    fv4[0] = 0.0f; fv4[1] = 0.0f; fv4[2] =  1.0f; fv4[3] = 0.0f;
    glLightfv(GL_LIGHT0, GL_POSITION, fv4);
	fv4[0] = 0.0f; fv4[1] = 0.0f; fv4[2] = -1.0f; fv4[3] = 0.0f;
	glLightfv(GL_LIGHT1, GL_POSITION, fv4);
    fv4[0] = 0.9f; fv4[1] = 0.9f; fv4[2] = 0.9f; fv4[3] = 1.0f;
    glLightfv(GL_LIGHT0, GL_DIFFUSE, fv4);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, fv4);
	fv4[0] = 1.0f; fv4[1] = 1.0f; fv4[2] = 1.0f; fv4[3] = 1.0f;
    glLightfv(GL_LIGHT0, GL_SPECULAR, fv4);
	glLightfv(GL_LIGHT1, GL_SPECULAR, fv4);
    
    glEnable(GL_LIGHTING);

		// Definizione colore base
    fv4[0] = 0.0f;
    fv4[1] = 0.0f;
    fv4[2] = 0.0f;
    fv4[3] = 1.0f;
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, fv4);
    iv1[0] = 0;
    glMaterialiv(GL_FRONT_AND_BACK, GL_SHININESS, iv1);
	
	
}

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView

IMPLEMENT_DYNCREATE(CGlwrapperView, CView)

BEGIN_MESSAGE_MAP(CGlwrapperView, CView)
	//{{AFX_MSG_MAP(CGlwrapperView)
	ON_WM_SIZE()
	ON_WM_ERASEBKGND()
	ON_WM_DESTROY()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	//}}AFX_MSG_MAP
	// Standard printing commands
	ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
	ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView construction/destruction

CGlwrapperView::CGlwrapperView()
{
	m_pOldPalette = NULL;
	m_pDC		  = NULL;
	m_hrc		  = NULL;

	m_TrackBall.SetSpin(FALSE);
		//Azzeramento matrici
	int i,j;
	for(j=0;j<4;j++)
		for(i=0;i<4;i++)
		{
			if(i==j) m_matRot[j][i] = 1.0f;
			else     m_matRot[j][i] = 0.0f;
		}

			// Azzeramento traslazione utente
	m_utrans[0] = 0.0f;
	m_utrans[1] = 0.0f;
	m_utrans[2] = 0.0f;
}

CGlwrapperView::~CGlwrapperView()
{
}

BOOL CGlwrapperView::PreCreateWindow(CREATESTRUCT& cs)
{
		// An OpenGL window must be created with the following flags and must not
		// include CS_PARENTDC for the class style. Refer to SetPixelFormat
		// documentation in the "Comments" section for further information.
    cs.style |= WS_CLIPSIBLINGS | WS_CLIPCHILDREN | CS_OWNDC;
	if(cs.style&CS_PARENTDC)
	{
		AfxMessageBox("Win OpenGL Internal Error!");
		return FALSE;
	}

	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView drawing

void CGlwrapperView::OnDraw(CDC* pDC)
{
	if(pDC==NULL) return;

	CGlwrapperDoc* pDoc = GetDocument();
	ASSERT_VALID(pDoc);
		// Semplice semaforo
	static BOOL bBusy = FALSE;
	if(bBusy)
		return;
	bBusy = TRUE;

		// Setta il contesto corrente
	SetGL();

		// Cancellazione buffer
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		// Il suo punto di vista
	glLoadIdentity();
	glTranslatef(m_utrans[0],m_utrans[1],m_utrans[2]);
	glMultMatrixf( &(m_matRot[0][0]) );
	
	
	glBegin(GL_TRIANGLES);
		glColor3f(0,0,1);
		glVertex3f( 0, 5,-10);
		glColor3f(0,1,0);
		glVertex3f(-5,-5,-10);
		glColor3f(1,0,0);
		glVertex3f( 5,-5,-10);
	glEnd();

	glFinish();
    if(pDC!=NULL)
		SwapBuffers(wglGetCurrentDC());

	bBusy = FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView printing

BOOL CGlwrapperView::OnPreparePrinting(CPrintInfo* pInfo)
{
	// default preparation
	return DoPreparePrinting(pInfo);
}

void CGlwrapperView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add extra initialization before printing
}

void CGlwrapperView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
	// TODO: add cleanup after printing
}

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView diagnostics

#ifdef _DEBUG
void CGlwrapperView::AssertValid() const
{
	CView::AssertValid();
}

void CGlwrapperView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CGlwrapperDoc* CGlwrapperView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CGlwrapperDoc)));
	return (CGlwrapperDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CGlwrapperView message handlers

void CGlwrapperView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);
	
	if(cx!=0 && cy!=0)
	{
		const float fNearPlane =  0.1f;
		const float fFarPlane  = 20.0f;
		const float fPAngle	   = 45.0f;

		SetGL();
		glViewport(0, 0, cx, cy);
		GLfloat fAspect;
		if (cy) fAspect = GLfloat(cx)/cy;
		else	fAspect = 1.0f;
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(fPAngle, fAspect, fNearPlane, fFarPlane);
		glMatrixMode(GL_MODELVIEW);

		m_TrackBall.Resize( cx,cy );
	}	
}

BOOL CGlwrapperView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;
}

void CGlwrapperView::OnDestroy() 
{
	SetGL();
	glFinish();
	wglMakeCurrent(NULL,  NULL);
    if (m_hrc) ::wglDeleteContext(m_hrc);
    if (m_pOldPalette) m_pDC->SelectPalette(m_pOldPalette, FALSE);
	if (m_pDC) delete m_pDC;
	m_glOwner = NULL;			// Si cancella da owner corrente
	
	CView::OnDestroy();	
}


BOOL CGlwrapperView::Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext) 
{
	
				// Creazione della finestra
	if( !CWnd::Create(lpszClassName, lpszWindowName, dwStyle,
		             rect, pParentWnd, nID, pContext) )
		return FALSE;

		// Creazione DC
	m_pDC = new CClientDC(this);
    ASSERT(m_pDC != NULL);

		// Setta il pixel format e crea la palette
	if (!SetupPixelFormat(m_pDC))
        return FALSE;
	
		// Creazione palette 
	if( CreateRGBPalette(m_pDC,m_cPalette) )
	{
		m_pOldPalette = m_pDC->SelectPalette(&m_cPalette, FALSE);
        m_pDC->RealizePalette();
	}

		// Crea e setta il contesto OPENGL
	m_hrc = wglCreateContext(m_pDC->GetSafeHdc());
	if(m_hrc==NULL)
	{
		AfxMessageBox("OpenGL contest fail");
		return FALSE;
	}

		// Inizializzazioni GL
	SetGL();
	InitGL();
	//glClearColor(m_BgColor[0],m_BgColor[1],m_BgColor[2],1.0f);

			// Init Viewew geometry
	CRect rc;
	GetClientRect(&rc);
	m_TrackBall.Init(rc.right,rc.bottom);
	
	return TRUE;

}

void CGlwrapperView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	if(!m_bCaptured)
	{
		m_TrackBall.SetSpin(FALSE);
		if( nFlags & MK_CONTROL )
			m_nMode = 1;
		else
			m_nMode = 0;
		m_TrackBall.MouseDown( point.x, point.y, m_nMode );
		SetCapture();
		m_bCaptured = TRUE;
	}
}

void CGlwrapperView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if(m_bCaptured)
	{
		m_TrackBall.MouseUp(point.x,point.y);
		ReleaseCapture();
		m_bCaptured = FALSE;
	}
}

void CGlwrapperView::OnMouseMove(UINT nFlags, CPoint point) 
{
	if(m_bCaptured)
	{
		if(m_nMode==0)
			m_TrackBall.CalcRotMatrix( m_matRot, point.x, point.y );
		else if(m_nMode==1)
			m_TrackBall.CalcScaVector( m_utrans, point.x, point.y );
		else if(m_nMode==2)
			m_TrackBall.CalcTraVector( m_utrans, point.x, point.y );

		if( (nFlags & MK_CONTROL)==0 )
		{
		if( (nFlags & MK_LBUTTON)!=0 && (nFlags & MK_MBUTTON)!=0)
		{
			if(m_nMode!=1)
			{
				m_TrackBall.MouseUp(point.x,point.y);
				m_nMode = 1;
				m_TrackBall.MouseDown( point.x, point.y, m_nMode );
			}
		}
		else if( (nFlags & MK_LBUTTON)!=0 )
		{
			if(m_nMode!=0 )
			{
				m_TrackBall.MouseUp(point.x,point.y);
				m_nMode = 0;
				m_TrackBall.MouseDown( point.x, point.y, m_nMode );
			}
		}
		else if( (nFlags & MK_MBUTTON)!=0 )
		{
			if( m_nMode!=2 )
			{
				m_TrackBall.MouseUp(point.x,point.y);
				m_nMode = 2;
				m_TrackBall.MouseDown( point.x, point.y, m_nMode );
			}
		}
		}

		Invalidate();
	}	
}
